@@ -2,6 +2,8 @@ import numpy as np
 from scipy import constants
 from typing import Tuple, List
 from math import degrees, atan
+from functools import lru_cache
+import tensorflow as tf
 
 pi = constants.pi  # define pi, the ratio of the circumference of a circle to its diameter
 g = constants.g    # define the gravitational acceleration on earth, in ms^-2
@@ -70,6 +72,14 @@ def fibonacci_sequence(n) -> List[int]:
     return fibo_sequence
 
 
+# a showcase of recurrence (in this case of course completely unnecessary) and function decorators
+# function values are stored, as in, for 10!, we have to calculate 9!. If we:
+# call f(10), and after that call f(9), then f(9) will not be calculated, as it is cached
+@lru_cache(maxsize=None)
+def factorial(n):
+    return n * factorial(n-1) if n else 1
+
+
 class EstimatePi:
 
     @staticmethod
@@ -115,4 +125,27 @@ class EstimatePi:
 
         error = abs(pi_estimation - pi)
 
-        return error, pi_estimation
\ No newline at end of file
+        return error, pi_estimation
+
+def functional_std_approximation(f, variables, size=1000):
+    """
+    Give a functional approximation of the standard deviation after performing a function on the inputs.
+    """
+    return np.std(f(*[np.random.normal(m, s, size=size) for m, s in variables]))
+
+def calculus_std_approximation(f, variables: List[Tuple[float]]):
+    """
+    Give an approximation of the standard deviation after performing a function on the inputs, using calculus.
+
+    Uses the autograd property of tensorflow to calculate the derivatives. 
+    """
+    tf_variables = [tf.Variable(x[0]) for x in variables]
+    # log the variables using GradientTape, so the derivatives can later be easily extracted
+    with tf.GradientTape() as tape:
+        y = f(*tf_variables)
+    derivatives = np.abs(np.array(tape.gradient(y, tf_variables)))
+
+    # add different derivatives using root of the sum of squares
+    relative_uncertainties = [(derivative * x[1])**2 for derivative, x in zip(derivatives, variables)]
+    uncertainty = np.sum(relative_uncertainties)**.5
+    return uncertainty
